import * as React from 'react';
import {
  Button,
  Content,
  Modal,
  ModalBody,
  ModalFooter,
  ModalHeader,
  TabTitleText,
} from '@patternfly/react-core';
import { WorkspaceRedirectInformationView } from '~/app/pages/Workspaces/workspaceActions/WorkspaceRedirectInformationView';
import { Workspace } from '~/shared/api/backendApiTypes';
import { ActionButton } from '~/shared/components/ActionButton';

interface StopActionAlertProps {
  onClose: () => void;
  isOpen: boolean;
  workspace: Workspace | null;
  onStop: () => Promise<void>;
  onUpdateAndStop: () => Promise<void>;
  onActionDone: () => void;
}

type StopAction = 'stop' | 'updateAndStop';

export const WorkspaceStopActionModal: React.FC<StopActionAlertProps> = ({
  onClose,
  isOpen,
  workspace,
  onStop,
  onUpdateAndStop,
  onActionDone,
}) => {
  const workspacePendingUpdate = workspace?.pendingRestart;
  const [actionOnGoing, setActionOnGoing] = React.useState<StopAction | null>(null);

  const executeAction = React.useCallback(
    async (args: { action: StopAction; callback: () => Promise<void> }) => {
      setActionOnGoing(args.action);
      try {
        return await args.callback();
      } finally {
        setActionOnGoing(null);
      }
    },
    [],
  );

  const handleStop = React.useCallback(async () => {
    try {
      await executeAction({ action: 'stop', callback: onStop });
      // TODO: alert user about success
      console.info('Workspace stopped successfully');
      onActionDone();
      onClose();
    } catch (error) {
      // TODO: alert user about error
      console.error('Error stopping workspace:', error);
    }
  }, [executeAction, onActionDone, onClose, onStop]);

  // TODO: combine handleStop and handleUpdateAndStop if they end up being similar
  const handleUpdateAndStop = React.useCallback(async () => {
    try {
      await executeAction({ action: 'updateAndStop', callback: onUpdateAndStop });
      // TODO: alert user about success
      console.info('Workspace updated and stopped successfully');
      onActionDone();
      onClose();
    } catch (error) {
      // TODO: alert user about error
      console.error('Error updating and stopping workspace:', error);
    }
  }, [executeAction, onActionDone, onClose, onUpdateAndStop]);

  const shouldShowActionButton = React.useCallback(
    (action: StopAction) => !actionOnGoing || actionOnGoing === action,
    [actionOnGoing],
  );

  return (
    <Modal
      variant="medium"
      isOpen={isOpen}
      aria-describedby="modal-title-icon-description"
      aria-labelledby="title-icon-modal-title"
      onClose={onClose}
    >
      <ModalHeader title="Stop Workspace" />
      <ModalBody>
        {workspacePendingUpdate ? (
          <>
            <TabTitleText>
              There are pending redirect updates for that workspace. Are you sure you want to
              proceed?
            </TabTitleText>
            <WorkspaceRedirectInformationView kind={workspace.workspaceKind.name} />
          </>
        ) : (
          <Content>Are you sure you want to stop the workspace?</Content>
        )}
      </ModalBody>
      <ModalFooter>
        {shouldShowActionButton('updateAndStop') && workspacePendingUpdate && (
          <ActionButton
            action="Update and Stop"
            titleOnLoading="Stopping ..."
            onClick={() => handleUpdateAndStop()}
          >
            Update and Stop
          </ActionButton>
        )}

        {shouldShowActionButton('stop') && (
          <ActionButton
            action="Stop"
            titleOnLoading="Stopping ..."
            onClick={() => handleStop()}
            variant={workspacePendingUpdate ? 'secondary' : 'primary'}
          >
            {workspacePendingUpdate ? 'Stop and defer updates' : 'Stop'}
          </ActionButton>
        )}

        {!actionOnGoing && (
          <Button variant="link" onClick={onClose}>
            Cancel
          </Button>
        )}
      </ModalFooter>
    </Modal>
  );
};
